;; Copyright 2011, 2012, 2013, 2014, 2015, 2016 Kevin Ryde

;; Author: Kevin Ryde <user42_kevin@yahoo.com.au>
;; Version: 0
;; Keywords: data, OEIS
;; URL: http://user42.tuxfamily.org/oeis-el/index.html
;;
;; oeis.el is free software; you can redistribute it and/or modify it under
;; the terms of the GNU General Public License as published by the Free
;; Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; oeis.el is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
;; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
;; more details.
;;
;; You can get a copy of the GNU General Public License online at
;; <http://www.gnu.org/licenses/>.



;;; Commentary:
;; 

;;; Code:

(eval-and-compile
  (unless (fboundp 'overlay-put)
    (require 'overlay))) ;; for xemacs21

(eval-when-compile
  (unless (and (fboundp 'declare)
               (fboundp 'dolist)
               (fboundp 'ignore-errors)
               (fboundp 'pop))
    (require 'cl)))

;;------------------------------------------------------------------------------
;; A-number to name

(defvar oeis-anum-to-name-cache nil
  "An internal part of oeis.el.")

(defun oeis-anum-to-name-cache-store (elem)
  "An internal part of oeis.el.
ELEM is a pair of strings (anum . name)."
  (setq oeis-anum-to-name-cache
        (cons elem
              (delete (assoc anum oeis-anum-to-name-cache)
                      oeis-anum-to-name-cache)))
  ;; limit its size
  (let ((tail (nthcdr 20 oeis-anum-to-name-cache)))
    (if tail (setcdr tail nil)))
  elem)

(eval-when-compile
  (defmacro oeis-with-find-file (filename &rest body)
    "An internal part of oeis.el.
This macro does not exist when running byte compiled.

Visit FILENAME and evaluate BODY.  If FILENAME is in an existing
buffer then use that, otherwise visit it temporarily and kill the
buffer after BODY.  Point is at the start of the buffer and is in
a`save-excursion' so can be modified.

If FILENAME does not exist then do not evaluate BODY and return
nil."

    (declare (debug t) (indent 1))
    `(let* ((oeis-with-find-file--filename ,filename)
            (oeis-with-find-file--existing-buffer
             (find-buffer-visiting oeis-with-find-file--filename))
            (oeis-with-find-file--buffer
             (or oeis-with-find-file--existing-buffer
                 (ignore-errors (find-file-noselect
                                 oeis-with-find-file--filename)))))
       (and oeis-with-find-file--buffer
            (unwind-protect
                (with-current-buffer oeis-with-find-file--buffer
                  (save-excursion
                    (goto-char (point-min))
                    ,@body))
              (unless oeis-with-find-file--existing-buffer
                (kill-buffer oeis-with-find-file--buffer)))))))

(defun oeis-anum-to-name-from-internal-text (anum)
  "An internal part of oeis.el.
Return the name of sequence ANUM from ~/OEIS/ANUM.internal.txt.
Return nil if no such file or somehow no %N line in the file."

  (let ((filename (oeis-local-filename (concat anum ".internal.txt"))))
    (when filename
      (let ((coding-system-for-read (if (memq 'utf-8 (coding-system-list))
                                        'utf-8 coding-system-for-read)))
        (oeis-with-find-file filename
          (when (re-search-forward "^%N[ \t]+" (regexp-quote anum) "[ \t]+" nil t)
            (buffer-substring (point) (line-end-position))))))))

(defun oeis-anum-to-name-from-internal-html (anum)
  "An internal part of oeis.el.
Return the name of sequence ANUM from ~/OEIS/ANUM.internal.html.
Return nil if no such file or somehow no %N line in the file.

The file may have HTML tags and \"&\" entities.  Tags are
stripped and entities are decoded for the return."

  (let ((filename (oeis-local-filename (concat anum ".internal.html"))))
    (when filename
      (oeis-with-find-file filename
        (when (re-search-forward "^[ \t]*\\(<[^>]*>\\)*[ \t]*%N +\\(.*\\)$" nil t)
          (let ((str (match-string 2)))
            (setq str (replace-regexp-in-string "<[^>]*>" "" str ;; <tags>
                                                t   ;; fixedcase
                                                t)) ;; literal
            (ignore-errors
              (require 'xml)
              ;; no xml-substitute-numeric-entities in xml.el of xemacs21
              (setq str (xml-substitute-numeric-entities str)))
            str))))))

(defun oeis-anum-to-name-from-names (anum)
  "An internal part of oeis.el.
Return the name of sequence ANUM from ~/OEIS/names or
~/OEIS/names.gz file.  Return nil if no names file or no such
ANUM."
  (let ((filename (or (oeis-local-filename "names")
                      (and auto-compression-mode
                           (oeis-local-filename "names.gz"))))
        (large-file-warning-threshold nil))
    (when filename
      (oeis-with-find-file filename
        (when (re-search-forward (concat "^" (regexp-quote anum) "[ \t]+")
                                 nil t)
          (buffer-substring (point) (line-end-position)))))))

(defun oeis-anum-to-name (anum)
  "Return the name of sequence ANUM, or return nil if unknown.
ANUM is a string like \"A000001\".  The name is sought in files

    ~/OEIS/ANUM.internal.txt       \"internal format\"
    ~/OEIS/names                   whole names file
    ~/OEIS/names.gz

The names file is a very big list of all A-number names.  Having
it makes all names available.  URL `http://oeis.org/names.gz'.
URL `http://oeis.org/allfiles.html' describes it and various
other files."

  (let ((name (or (cdr (assoc anum oeis-anum-to-name-cache))
                  (oeis-anum-to-name-from-internal-html anum)
                  (oeis-anum-to-name-from-names anum))))
    (oeis-anum-to-name-cache-store (cons anum name)) ;; as least recently used
    name))

(defun oeis-display-name-at-point ()
  (interactive)
  (require 'oeis)
  (let* ((anum (oeis-anum-at-point))
         (name (and anum (oeis-anum-to-name anum))))
    (cond (name
           (message "%s %s" anum name))
          (anum
           (message "%s unknown" anum))
          (t
           (message "No A-number at point" anum)))))

;;-----------------------------------------------------------------------------
;; tooltip minor mode

(defun oeis-tooltip-help-echo-handler (arg &rest args)
  "An internal part of oeis.el.
This is designed for use as in a `help-echo' property.
OVERLAY is on buffer text like \"A123456\" or \"b123456.txt\".
Return a string which is the name of this sequence, or return nil
if no name known.

The A-number is included in the returned string, as a way to
confirm to the user what the description refers to.  If a few
A-numbers are close together then you might not be certain quite
which the mouse is over, and the tooltip may or may not update
quickly."

  (message "oeis-tooltip-help-echo-handler %S %S" arg args)
  ;; Emacs  (WINDOW OVERLAY POS)
  ;; XEmacs (OVERLAY)
  (let* ((overlay (if (zerop (length args)) arg (car args)))
         (str     (with-current-buffer (overlay-buffer overlay)
                    (buffer-substring-no-properties (overlay-start overlay)
                                                    (overlay-end overlay)))))
    ;; (message "oeis-tooltip-help-echo-handler on %S" str)
    (when (string-match "\\([0-9]+\\)" str)
      (setq str (concat "A" (match-string 0 str)))
      (let ((name (oeis-anum-to-name str)))
        (concat str " " name)))))

(defun oeis-tooltip-idle-handler ()
  "An internal part of oeis.el.
An idle timer callback which adds `help-echo' overlays to
    A123456
    a123456.txt
    b123456.txt
showing the sequence name."

  (setq oeis-tooltip-idle-timer nil)
  (walk-windows
   (lambda (window)
     (with-current-buffer (window-buffer window)
       (save-excursion
         (goto-char (window-end window t))
         (let ((limit (line-end-position))
               overlays)
           (goto-char (window-start window))
           (beginning-of-line)
           (setq overlays (overlays-in (point) limit))

           ;; (message "update %S to %S" (point) limit)

           (while (re-search-forward
                   "\\b\\(A[0-9]\\{6,7\\}\\|[ab][0-9]\\{6,7\\}\\.txt\\)\\b"
                   limit t)
             (let ((beg (match-beginning 0))
                   (end (match-end 0))
                   overlay)
               ;; move an existing overlay of ours, otherwise create a new
               (while (and (setq overlay (pop overlays))
                           (not (eq (overlay-get overlay 'help-echo)
                                    'oeis-tooltip-help-echo-handler))))
               (if overlay
                   (move-overlay overlay beg end)
                 (overlay-put (make-overlay beg end (current-buffer) nil nil)
                              'help-echo 'oeis-tooltip-help-echo-handler))))

           ;; Remove any leftover overlays of ours, perhaps due to deleted
           ;; text.
           ;;
           ;; emacs21 and xemacs21 don't have `remove-overlays' (new in
           ;; emacs22), but the delere here is not the same as
           ;; `remove-overlays' anyway.  `remove-overlays' splits overlays
           ;; crossing the endpoints whereas here want to delete completely.
           ;;
           (dolist (overlay overlays)
             (if (eq (overlay-get overlay 'help-echo)
                     'oeis-tooltip-help-echo-handler)
                 (delete-overlay overlay)))))))

   nil          ;; not the minibuffer
   'visible))   ;; all windows in all visible frames

(defvar oeis-tooltip-idle-timer nil
  "An internal part of oeis.el.
An idle timer object which will call `oeis-tooltip-idle', or nil
if no such timer running.")

(defun oeis-tooltip-start-timer ()
  "An internal part of oeis.el.
Start an idle timer to call `oeis-tooltip-idle', if not already running.
The timer is held in `oeis-tooltip-idle-timer'."
  (unless oeis-tooltip-idle-timer
    (setq oeis-tooltip-idle-timer
          (run-with-idle-timer .1    ;; seconds
                               nil   ;; no repeat, once only
                               'oeis-tooltip-idle-handler))))

(defun oeis-tooltip-window-configuration-change ()
  "An internal part of oeis.el.
This is designed for use in 'window-configuration-change-hook'.
Start an idle timer to update OEIS tooltip overlays."
  (oeis-tooltip-start-timer))

(defun oeis-tooltip-window-scroll-function (window pos)
  "An internal part of oeis.el.
This is designed for use in 'window-scroll-functions'.
Start an idle timer to update OEIS tooltip overlays."
  (oeis-tooltip-start-timer))

(defun oeis-tooltip-after-change (beg end prev-len)
  "An internal part of oeis.el.
This is designed for use in `after-change-functions'.
If the changed buffer is visible then start an idle timer to
update OEIS tooltip overlays."
  (if (get-buffer-window (current-buffer)
                         'visible) ;; all windows on all visible frames
      (oeis-tooltip-start-timer)))

;; (defun oeis-tooltip-post-command ()
;;   "An internal part of oeis.el.
;; This is designed for use in 'post-command-hook'.
;; Start an idle timer to update OEIS tooltip overlays."
;;   (oeis-tooltip-start-timer))

(define-minor-mode oeis-tooltip-mode
  "Add help-echo tooltip to A-numbers."
  :type   'boolean
  :global t

  (if oeis-tooltip-mode
      (progn
        (add-hook 'after-change-functions
                  'oeis-tooltip-after-change)
        (add-hook 'window-scroll-functions
                  'oeis-tooltip-window-scroll-function)
        (add-hook 'window-configuration-change-hook
                  'oeis-tooltip-window-configuration-change)
        ;; (add-hook 'post-command-hook
        ;;           'oeis-tooltip-post-command)
        (oeis-tooltip-idle-handler)) ;; initial overlay creation

    ;; disable
    (remove-hook 'after-change-functions
                 'oeis-tooltip-after-change)
    (remove-hook 'window-scroll-functions
                 'oeis-tooltip-window-scroll-function)
    (remove-hook 'window-configuration-change-hook
                 'oeis-tooltip-window-configuration-change)
    ;; (remove-hook 'post-command-hook
    ;;              'oeis-tooltip-post-command)

    ;; Delete all `oeis-tooltip-help-echo-handler' overlays everywhere.
    ;; Perhaps iterating `next-overlay-change' would be more efficient than
    ;; asking for overlays-in whole buffer.
    (dolist (buffer (buffer-list))
      (with-current-buffer buffer
        (save-restriction
          (widen)
          (dolist (overlay (overlays-in (point-min) (point-max)))
            (if (eq (overlay-get overlay 'help-echo)
                    'oeis-tooltip-help-echo-handler)
                (delete-overlay overlay))))))))

(defun oeis-tooltip-enable ()
  (interactive)
  (oeis-tooltip-mode 1))


(provide 'oeis-tooltip)

;;; oeis-tooltip.el ends here
