;; Copyright 2009, 2010, 2011, 2013, 2017 Kevin Ryde

;; This file is part of miscbits-el.
;;
;; miscbits-el is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the
;; Free Software Foundation; either version 3, or (at your option) any later
;; version.
;;
;; miscbits-el is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
;; Public License for more details.
;;
;; You should have received a copy of the GNU General Public License along
;; with miscbits-el; see the file COPYING.  Failing that, go to
;; <http://www.gnu.org/licenses/>.

(add-to-list 'load-path (expand-file-name ".."))

(defun my-string-make-unibyte (str)
  (if (fboundp 'string-make-unibyte)
      (string-make-unibyte str)
    str))

(defconst my-unicode-bullet-str
  (or (condition-case nil
          (string (decode-char 'ucs 8226)) ;; U+2022
        (error nil))
      (condition-case nil
          (decode-coding-string (my-string-make-unibyte (string 226 128 162))
                                'utf-8)
        (error nil))))

(defun my-run-test ()

  (let ((count 0)
        (tempbuf "*t-ffap-href--temp*")
        (case-fold-search t))
    (switch-to-buffer tempbuf)

    ;;-------------------------------------------------------------------------
    ;; autoloads

    (require 'miscbits-el-loaddefs)

    (mapcar (lambda (func)
              (unless (fboundp func)
                (error "`%S' not autoloaded" func))
              (setq count (1+ count)))
            '(ffap-href-enable))

    (mapcar (lambda (func)
              (unless (commandp func)
                (error "`%S' not interactive" func))
              (setq count (1+ count)))
            '(ffap-href-enable))

    ;;-------------------------------------------------------------------------
    ;; load

    (require 'ffap)
    (require 'ffap-href)

    ;; no leftovers in emacs21 byte compiling defadvice
    (when (boundp 'byte-compile-unresolved-functions)
      (setq byte-compile-unresolved-functions
            (delete '(thing-at-point-looking-at 1) ;; loaded dynamically
                    byte-compile-unresolved-functions))
      (when byte-compile-unresolved-functions
        ;; (message "unr %S" byte-compile-unresolved-functions)
        ;; (message "fb %S" (fboundp (caar byte-compile-unresolved-functions)))
        (error "byte-compile-unresolved-functions after load: %S"
               byte-compile-unresolved-functions)))
    (setq count (1+ count))

    ;;-------------------------------------------------------------------------
    ;; ffap-url-at-point, ffap-file-at-point -- when never enabled

    (ad-activate 'ffap-url-at-point)
    (ad-activate 'ffap-file-at-point)

    (erase-buffer)
    (let ((got  (ffap-url-at-point))
          (want nil))
      (unless (equal got want)
        (error "ffap-url-at-point in empty buffer got: %S" got))
      (setq count (1+ count)))
    (let ((got  (ffap-file-at-point))
          (want nil))
      (unless (equal got want)
        (error "ffap-file-at-point in empty buffer got: %S" got))
      (setq count (1+ count)))

    (erase-buffer)
    (insert "<a href=\"foo/subdir/index.html\">blah</a>")
    (goto-char (point-min))

    (let ((got  (ffap-url-at-point))
          (want nil))
      (unless (equal got want)
        (error "ffap-url-at-point when never enabled got: %S" got))
      (setq count (1+ count)))

    (let ((got  (ffap-file-at-point))
          (want nil))
      (unless (equal got want)
        (error "ffap-file-at-point when never enabled got: %S" got))
      (setq count (1+ count)))


    ;;-------------------------------------------------------------------------
    ;; ffap-url-at-point, ffap-file-at-point -- enable

    (erase-buffer)
    (insert "<a href=\"foo/subdir/index.html\">blah</a>")
    (goto-char (point-min))

    (ffap-href-enable)

    (let ((got  (ffap-file-at-point))
          (want "foo/subdir/index.html"))
      (unless (equal got want)
        (error "ffap-file-at-point when enabled got: %S" got))
      (setq count (1+ count)))

    (erase-buffer)
    (insert "<a href=\"http://foo.org/index.html\">blah</a>")
    (goto-char (point-min))

    (let ((got  (ffap-url-at-point))
          (want "http://foo.org/index.html"))
      (unless (equal got want)
        (error "ffap-url-at-point when enabled got: %S" got))
      (setq count (1+ count)))

    ;;-------------------------------------------------------------------------
    ;; ffap-href-ucs-string

    (mapcar (lambda (elem)
              (let* ((num  (nth 0 elem))
                     (want (nth 1 elem))
                     (got  (ffap-href-ucs-string num)))
                (unless (equal got want)
                  (error "ffap-href-ucs-string %S got %S want %S"
                         num got want))
                (setq count (1+ count))))

            `((88   "X")
              (247  "") ;; division sign
              (8226 ,my-unicode-bullet-str) ;; or nil if unavailable
              ))

    ;;-------------------------------------------------------------------------
    ;; ffap-href-unentitize

    (mapcar (lambda (elem)
              (let* ((str  (nth 0 elem))
                     (want (nth 1 elem))
                     (got  (ffap-href-unentitize str)))
                (unless (equal got want)
                  (error "ffap-href-unentitize %S got %S want %S"
                         str got want))
                (setq count (1+ count))))

            `(("&lt;"     "<")
              ("&amp;lt;" "&lt;")
              ("&#88;"    "X")
              ("&#247;"   "") ;; division sign

              ("&#8226;"  ,(or my-unicode-bullet-str ;; bullet if available
                               "&#8226;"))
              ))


    ;;-------------------------------------------------------------------------
    ;; ffap-href-at-point

    (mapcar (lambda (elem)
              (let* ((str  (nth 0 elem))
                     (want (nth 1 elem)))
                (erase-buffer)
                (insert str)
                (goto-char (point-min))

                (let ((got (ffap-href-at-point)))
                  (unless (equal got want)
                    (error "ffap-href-at-point %S got %S want %S"
                           str got want)))
                (setq count (1+ count))))

            `(("<a href=\"../blah.html\">"  "../blah.html")
              ("<a href='http://localhost/index.html'>"  "http://localhost/index.html")
              ("<a href='http://localhost/inde&#88;.html'>"
               "http://localhost/indeX.html")))


    ;;-------------------------------------------------------------------------
    ;; unload-feature

    (add-to-list 'load-path (expand-file-name "."))
    (require 'my-unload)

    (if (not my-unload-feature-function-p)
        (message "skip, no unload-feature function")

      (let ((feature 'ffap-href))
        (unless (my-advice-by-name feature)
          (error "oops, didn't find any advice"))

        (unload-feature feature t)

        (let ((lst (my-advice-by-name feature)))
          (when lst
            (error "unload-feature leftover advice: %S" lst)))
        (setq count (1+ count))))

    ;;-------------------------------------------------------------------------
    ;; unload-feature -- with advice unloaded first

    (let ((feature 'ffap-href))
      (if (featurep feature)
          (unload-feature feature t))
      (require feature)
      ;; if 'advice can be unloaded (no dependency expressed on it), then
      ;; see FEATURE unloads successfully too
      (when (condition-case nil
                (progn (unload-feature 'advice) t)
              (error nil))
        (unload-feature feature t)))
    (setq count (1+ count))

    ;;-------------------------------------------------------------------------
    ;; ffap-url-at-point, ffap-file-at-point -- when unloaded

    (require 'ffap-href)
    (ffap-href-unload-function)

    (erase-buffer)
    (insert "<a href=\"foo/subdir/index.html\">blah</a>")
    (goto-char (point-min))

    (let ((got  (ffap-url-at-point))
          (want nil))
      (unless (equal got want)
        (error "ffap-url-at-point when disabled got: %S" got))
      (setq count (1+ count)))

    (let ((got  (ffap-file-at-point))
          (want nil))
      (unless (equal got want)
        (error "ffap-file-at-point when disabled got: %S" got))
      (setq count (1+ count)))

    ;;-------------------------------------------------------------------------
    ;; ffap-url-at-point -- after unload-feature

    (if (featurep 'ffap-href)
        (unload-feature 'ffap-href))
    (let ((want"http://foo.org/index.html"))
      (erase-buffer)
      (insert want)
      (let ((got (ffap-url-at-point)))
        (unless (equal got want)
          (error "ffap-url-at-point after unload-feature:\ngot %S" got))
        (setq count (1+ count))))


    ;;-------------------------------------------------------------------------
    (bury-buffer)
    (message "  %d tests" count)
    count))
