;; cleanup when died?

;; return "spam" everywhere?


;;; kevspam.el --- gnus mail splitting with spamassassin in a subprocess.

;; Copyright 2002, 2003 Kevin Ryde
;;
;; kevspam.el is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by the
;; Free Software Foundation; either version 2, or (at your option) any later
;; version.
;;
;; kevspam.el is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
;; Public License for more details.
;;
;; You can get a copy of the GNU General Public License online at
;; http://www.gnu.org/licenses/gpl.txt, or you should have one in the file
;; COPYING which comes with GNU Emacs and other GNU programs.  Failing that,
;; write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.


;;; Commentary:

;; This code runs spamassassin as an Emacs sub-process and pipes messages to
;; it to be identified as spam or ham.
;;
;; This is similar to what spamc/spamd does, but doesn't fork for every
;; message.  On the authors system this ran much much faster, avoiding lots
;; of thrashing apparently caused by the forks.  Perhaps spamd will have an
;; option along these lines one day, rendering this package unnecessary.

;;; Usage:

;; Put the auxiliary kevspam.el somewhere in your lisp load-path, and
;; kevspam.pl somewhere in your normal execution PATH, then in .gnus
;;
;;    (autoload 'kevspam-split "kevspam" nil t)
;;    (setq nnmail-split-fancy
;;          '(|
;;            (: kevspam-split)
;;            ...
;;            ))
;;
;; If you want to send spam somewhere other than a group called "spam", then
;; make your own little function in the style of `kevspam-split'.  You can
;; test arbitrary message programmatically with `kevspam-test-str' if
;; desired.
;;
;; If you get some mail which has already been spam tagged, for instance a
;; list with good filtering, then it's worth matching that before
;; kevspam-split, since spamassassin is not especially fast, even when
;; already warmed up.
;;
;; CAUTION: You're strongly advised not to use the results of these tests to
;; actually delete mail.  I don't want anyone complaining that a bug in the
;; code lost them two dozen mails.  You're also advised to use set
;; mail-source-delete-incoming to nil.  And remember if something severe
;; goes wrong during splitting mail it should be left in
;; ~/.emacs-mail-crash-box (mail-source-crash-box).


;;; History:

;; Version 1 - the first version.


;;; Code:

(defvar kevspam-process nil
  "Process object for subordinate kevspam.pl.")

(defvar kevspam-process-buffer nil
  "Process buffer for subordinate kevspam.pl.")

(defun kevspam-cleanup ()
  "Cleanup any running kevspam process."
  (when kevspam-process
    (let ((proc kevspam-process))
      (setq kevspam-process nil)
      (kill-process proc))))
(add-hook 'gnus-exit-gnus-hook 'kevspam-cleanup)

(defun kevspam-test-str (str)
  "Return t if STR is spam."

  ;; start or restart the sub-process
  (setq kevspam-process-buffer (get-buffer-create "*kevspam*"))
  (when (or (not kevspam-process)
            (not (eq 0 (process-exit-status kevspam-process))))
    (message "Starting kevspam.pl")
    (setq kevspam-process
          (start-process "kevspam" kevspam-process-buffer "kevspam.pl"))
    (set-process-coding-system kevspam-process 'no-conversion 'no-conversion))

  ;; write to process
  ;; force to unibyte so we can be certain of the byte count
  (setq str (string-make-unibyte str))
  (process-send-string kevspam-process
                       (format "kevspam begin %d\n" (length str)))
  (process-send-string kevspam-process str)

  ;; look for "spam" or "good"
  (with-current-buffer kevspam-process-buffer
    (erase-buffer)
    (accept-process-output kevspam-process)
    (equal "spam\n"
           (buffer-substring (point-min)
                             (marker-position
                              (process-mark kevspam-process))))))

;;;###autoload
(defun kevspam-split ()
  "Test the current gnus split message and return \"spam\" or nil.

This function is intended for use in an `nnmail-split-fancy' rule, as for
instance

    (: kevspam-split)

to direct spam to the \"spam\" group, and leave ham for further split rules."

  (if (kevspam-test-str
       (save-excursion
         (save-restriction
           (widen)
           (buffer-substring (point-min) (point-max)))))
      "spam"
    nil))


(provide 'kevspam)

;;; kevspam.el ends here
